#ifndef IN_SCCS_ID
#ident " @(#)tcp.h (TWG) 2.7.1.2 88/11/11 ";
#endif /* IN_SCCS_ID */
/*
 * @(#) Copyright 1986.  The Wollongong Group, Inc.  All Rights Reserved.
 */
/*
 * Definitions of the TCP timers.  These timers are counted
 * down PR_SLOWHZ times a second.
 */
#define	TCPT_NTIMERS	4

#define	TCPT_REXMT	0		/* retransmit */
#define	TCPT_PERSIST	1		/* retransmit persistance */
#define	TCPT_KEEP	2		/* keep alive */
#define	TCPT_2MSL	3		/* 2*msl quiet time timer */

#define TCP_ALPHA	7
#define TCP_BETA	16

/*
 * The TCPT_REXMT timer is used to force retransmissions.
 * The TCP has the TCPT_REXMT timer set whenever segments
 * have been sent for which ACKs are expected but not yet
 * received.  If an ACK is received which advances tp->snd_una,
 * then the retransmit timer is cleared (if there are no more
 * outstanding segments) or reset to the base value (if there
 * are more ACKs expected).  Whenever the retransmit timer goes off,
 * we retransmit all unacknowledged segments, and do an exponential
 * backoff on the retransmit timer.
 *
 * The TCPT_PERSIST timer is used to keep window size information
 * flowing even if the window goes shut.  If all previous transmissions
 * have been acknowledged (so that there are no retransmissions in progress),
 * and the window is shut, then we start the TCPT_PERSIST timer, and at
 * intervals send a single byte into the peers window to force him to update
 * our window information.  We do this at most as often as TCPT_PERSMIN
 * time intervals, but no more frequently than the current estimate of
 * round-trip packet time.  The TCPT_PERSIST timer is cleared whenever
 * we receive a window update from the peer.
 *
 * The TCPT_KEEP timer is used to keep connections alive.  If an
 * connection is idle (no segments received) for TCPTV_KEEP amount of time,
 * but not yet established, then we drop the connection.  If the connection
 * is established, then we force the peer to send us a segment by sending:
 *	<SEQ=SND.UNA-1><ACK=RCV.NXT><CTL=ACK>
 * This segment is (deliberately) outside the window, and should elicit
 * an ack segment in response from the peer.  If, despite the TCPT_KEEP
 * initiated segments we cannot elicit a response from a peer in TCPT_MAXIDLE
 * amount of time, then we drop the connection.
 */

#define	TCP_TTL		30		/* time to live for TCP segs */
/*
 * Time constants.
 */
#define PR_SLOWHZ	2			/* ticks/second */
#define	TCPTV_MSL	1			/* max seg lifetime */
#define TCPTV_SRTTBASE	0			/* base roundtrip time;
						   if 0, no idea yet */
#define TCPTV_SRTTDFLT  (15 * PR_SLOWHZ)	/* assumed RTT if no info */
#define	TCPTV_KEEP	( 45*PR_SLOWHZ)		/* keep alive - 45 secs */
#define	TCPTV_PERSMIN	(  5*PR_SLOWHZ)		/* retransmit persistance */

#define	TCPTV_MAXIDLE	(8*TCPTV_KEEP)		/* maximum allowable idle
						   time before drop conn */
#define	TCPTV_MAXRXT	12			/* max times of retransmit */
#define	TCPTV_MAXKEEP	10			/* max times of missing keep-
						   alive before drop conn */

#define	TCPTV_MIN	(  1*PR_SLOWHZ)		/* minimum allowable value */
#define	TCPTV_MAX	( 30*PR_SLOWHZ)		/* maximum allowable value */
#define	TCPTV_REXMT	2			/* retransmit time value */
						/* should be adaptive */
#define	TCPTV_PERSIST	10			/* randomly chosen */

/*
 * Force a time value to be in a certain range.
 */
#define	TCPT_RANGESET(tv, value, tvmin, tvmax) { \
	(tv) = (value); \
	if ((tv) < (tvmin)) \
		(tv) = (tvmin); \
	if ((tv) > (tvmax)) \
		(tv) = (tvmax); \
}

typedef	unsigned long	tcp_seq;
/*
 * TCP header.
 * Per RFC 793, September, 1981.
 */

#ifdef XENIX
/* #pragma pack(1) */
#endif

struct tcphdr {
	ushort	th_sport;		/* source port */
	ushort	th_dport;		/* destination port */
	tcp_seq	th_seq;			/* sequence number */
	tcp_seq	th_ack;			/* acknowledgement number */
#if ( __vax__ || __ns32000__ || __i386__ )
	unchar	th_x2:4,		/* (unused) */
		th_off:4;		/* data offset */
#else
	unchar	th_off:4,		/* Data Offset	*/
		th_x2:4;		/* Unused	*/
#endif /* __vax__ */
	unchar	th_flags;
#define	TH_FIN	0x01
#define	TH_SYN	0x02
#define	TH_RST	0x04
#define	TH_PUSH	0x08
#define	TH_ACK	0x10
#define	TH_URG	0x20
	ushort	th_win;			/* window */
	ushort	th_sum;			/* checksum */
	ushort	th_urp;			/* urgent pointer */
};

#define	TCPHLEN		sizeof(struct tcphdr)

#define	TCPOPT_EOL	0
#define	TCPOPT_NOP	1
#define	TCPOPT_MAXSEG	2

/*
 * Tcp+ip header, after ip options removed.
 */
struct tcpiphdr {
	struct 	ipovly ti_i;		/* overlaid ip structure */
	struct	tcphdr ti_t;		/* tcp header */
};

#ifdef XENIX
/* #pragma pack() */
#endif

#define	ti_next		ti_i.ih_next
#define	ti_prev		ti_i.ih_prev
#define	ti_x1		ti_i.ih_x1
#define	ti_pr		ti_i.ih_pr
#define	ti_len		ti_i.ih_len
#define	ti_src		ti_i.ih_src
#define	ti_dst		ti_i.ih_dst
#define	ti_sport	ti_t.th_sport
#define	ti_dport	ti_t.th_dport
#define	ti_seq		ti_t.th_seq
#define	ti_ack		ti_t.th_ack
#define	ti_x2		ti_t.th_x2
#define	ti_off		ti_t.th_off
#define	ti_flags	ti_t.th_flags
#define	ti_win		ti_t.th_win
#define	ti_sum		ti_t.th_sum
#define	ti_urp		ti_t.th_urp

#define	TCPIPHLEN	sizeof(struct tcpiphdr)

/* from tcp_seq.h */

/*
 * TCP sequence numbers are 32 bit integers operated
 * on with modular arithmetic.  These macros can be
 * used to compare such integers.
 */
#define	SEQ_LT(a,b)	((int)((a)-(b)) < 0)
#define	SEQ_LEQ(a,b)	((int)((a)-(b)) <= 0)
#define	SEQ_GT(a,b)	((int)((a)-(b)) > 0)
#define	SEQ_GEQ(a,b)	((int)((a)-(b)) >= 0)

/*
 * Macros to initialize tcp sequence numbers for
 * send and receive from initial send and receive
 * sequence numbers.
 */
#define	tcp_rcvseqinit(tp) \
	(tp)->rcv_nxt = (tp)->irs + 1

#define	tcp_sendseqinit(tp) \
	(tp)->snd_una = (tp)->snd_nxt = (tp)->snd_max = (tp)->snd_up = \
	    (tp)->iss

#define	TCP_ISSINCR	128		/* increment for tcp_iss each second */

extern tcp_seq	tcp_iss;		/* tcp initial send seq # */

/*
 * tcp per channel structure for the upper queue
 */
struct tcpd {
	queue_t *td_rdq;		/* read queue */
	mblk_t	*td_tcpcb;		/* to protocol cntl block */
	struct	tcpd	*td_next;	/* The next device in active list */
	int	td_wcount;		/* # of bytes waiting to go down */
	int	td_rcount;		/* # of bytes waiting to go up */
	int	td_iocid;		/* Ioctl id's passed down.	*/
	mblk_t	*td_ipopt;		/* outgoing ip options */
	mblk_t	*td_ipoptin;		/* incoming ip options */
	short	td_topid;		/* who is on top */
	ushort 	td_sockflags;		/* socket option flags */
	unchar	td_flags;		/* see below */
	unchar	td_dev;			/* minor dev # */
	unchar	td_prot;		/* protocol # on top for demux */
	unchar	td_tstate;		/* TLI state */
	ushort	td_family;		/* AFUN - Address family */
};

/*
 * td_sockflags are the socket option flags (1<<0 thru 1<<8)
 * as defined in socket.h
 *
 * td_flags are as follows:
 */
#define	TD_ERROR	(1<<0)		/* fatal error occurred		*/
#define TD_SUPERUSER	(1<<1)		/* Super User opened it		*/
#define TD_RDWR		(1<<2)		/* Don't send M_PROTO for data	*/
#define TD_NOLINGER	(1<<3)		/* Connections linger		*/
#define TD_LINGER	~TD_NOLINGER
#define TD_CLOSING      (1<<4)          /* The user has issued a close  */
#define TD_NEWPORT      (1<<5)          /* This means he needs  new port*/


/*
 * tcp per channel structure for the bottom queue (mux)
 */
struct tcpb {
	queue_t	*tb_wrq;		/* write queue */
	int	tb_cookie;		/* cookie from LINK/UNLINK */
	char	tb_mstate;		/* mux state */
	short	tb_botid;		/* who is down below */
	mblk_t	*tb_iocblk;		/* temp. holding the M_IOCTL */
	queue_t *tb_qtop;		/* control chan write q */
};

#ifdef notdef
/*
 * ioctl format to the tcp module
 * should be accessible to user
 */
struct tcpioctl {
	int cookie;			/* from mux */
};
#endif /* notdef */

/*
 * from a pointer to tcpd, find the dev #
 * could have been TCPCHAN(tp) ((int)tp - (int)tcpd) / sizeof(struct tcpd))
 */
#define TCPCHAN(tp)	tp->td_dev
#define TCPCTLCHAN	0	/* tcp control channel */

/*
 * they should be somewhere else
 */
#define MUXSID	-2		/* sid for sadlog */
#define	UNREACH		255	/* unreachable td_tstate */
#define	SAMETLIST	254	/* same tli state */

#if __i386__
extern struct qhead tcpcb;
#else
extern struct qhead tcb;
#endif /* __i386__ */
extern struct tcpd tcpd[];
extern struct tcpb tcpb;


/*
 * Tcp control block, one per active tcp (after binding)
 * pointer back to tcpd (tcpd is alloc when the stream is opened)
 */
struct tcpcb {
	struct	tcpd *t_tdp;		/* back pointer to tcpd */
	struct	tsap t_la;		/* local address */
#define	t_laddr	t_la.ta_addr
#define	t_lport	t_la.ta_port
	struct	tsap t_fa;		/* foreign address */
#define	t_faddr	t_fa.ta_addr
#define	t_fport	t_fa.ta_port
	struct	qhead	t_conq;		/* conn queue */
	long	t_cqlen;		/* # of indications waiting in conq */
	long	t_maxcon;		/* max # of indications allowed*/
	long	t_conid;		/* conn id when floating */
	mblk_t	*t_template;		/* skeletal packet for transmit */
	struct	ipb	*t_ipb;		/* Pointer that is saved not used */
	struct	qhead	t_rsq;		/* reassembly queue */
#define	t_nseg	t_rsq.b_next
#define	t_pseg	t_rsq.b_prev
	short	t_state;		/* state of this connection */
	ushort	t_flags;
#define	TF_ACKNOW	0x01
#define	TF_FORCE	0x02
#define TF_DELACK	0x04		/* Delay the ACK	*/
#define TF_NOOPT	0x08
#define TF_SENTFIN	0x10		/* have sent fin	*/
	short	t_timer[TCPT_NTIMERS];	/* tcp timers */
	mblk_t	*t_tcpopt;		/* tcp options */
	ushort	t_maxseg;		/* maximum segment size */
/*
 * The following fields are used as in the protocol specification.
 * See RFC783, Dec. 1981, page 21.
 */
/* send sequence variables */
	ulong	snd_una;		/* send unacknowledged */
	ulong	snd_nxt;		/* send next */
	ulong	snd_up;			/* send urgent pointer */
	ulong	snd_wl1;		/* window update seg seq number */
	ulong	snd_wl2;		/* window update seg ack number */
	ulong	iss;			/* initial send sequence number */
	ushort	snd_wnd;		/* send window */
/* receive sequence variables */
	ushort	rcv_wnd;		/* receive window */
	ulong	rcv_nxt;		/* receive next */
	tcp_seq	rcv_adv;		/* Advertised window		*/
	ulong	rcv_up;			/* receive urgent pointer */
	ulong	irs;			/* initial rcv seq number */
/* retransmit variables */
	ulong	snd_max;		/* highest sequence number sent */
	ushort	t_rxt;			/* retransmit times 		*/
	ushort	t_rtt;			/* round trip timer		*/
	tcp_seq t_rtseq;		/* Sequence number being timed  */
	ushort	t_srtt;			/* Smoothed ROund trip timer	*/
	ushort	t_keep;			/* # of unack keep-alive mesg	*/
/* transmit timing stuff */
	ushort	t_idle;			/* inactivity time */
	ushort  t_srtv;			/* Smoothed Round Trip Variance */
/* source quench */
	ushort  snd_cwnd;		/* congestion controlled window */
	mblk_t	*t_lafu;		/* AFUN - Local afun address    */
	mblk_t	*t_fafu;		/* AFUN - Foreign afun address  */
};

#define	TILOG(x)	{errcode = x;}
/*
 * tcp error code in tcp_input.c
 */
#define	TCPE_HDR	1	/* tcp header error */
#define	TCPE_SUM	2	/* checksum */
#define	TCPE_OFFSET	3	/* bad offset */
#define	TCPE_OPTLEN	4	/* tcp option length */
#define	TCPE_ALLOCB	5	/* allocb fails */
#define	TCPE_LOOKUP	6	/* can't match a tcpcb */
#define	TCPE_BASE	7	/* db_base != b_rptr */
#define	TCPE_SSACK	10	/* SYN_SENT, got bad ack */
#define	TCPE_SSRST	11	/* SYN_SENT, got rst */
#define	TCPE_SSSYN	12	/* SYN_SENT, no syn */
#define	TCPE_SSADJ	13	/* SYN_SENT, adjmsg */
#define	TCPE_RWIN0	21	/* rcv_wnd=0, rcv_nxt != ti_seq */
#define	TCPE_OUTSEQ	22	/* rcv seg not in seq */
#define	TCPE_RSTRST	25	/* rcv rst, send rst */
#define	TCPE_RSTDROP	26	/* rcv rst, don't send rst, just drop */
#define	TCPE_BADSYN	27	/* rcv unexpected syn, close user, rst peer */
#define	TCPE_NOACK	31	/* TH_ACK is not set, drop seg */
#define	TCPE_SRBADACK	32	/* SYN_RCVD, bad ack */
#define	TCPE_BADACK	33	/* other than SYN_RCVD, bad ack */
#define	TCPE_LASTACK	35	/* get FIN|ACK in LAST_ACK */
#define	TCPE_CLOSING	36	/* get FIN|ACK in CLOSING */
#define	TCPE_TWAIT	37	/* get FIN|ACK in TIME_WAIT */
#define	TCPE_NQRST	41	/* new-req: has rst */
#define	TCPE_NQACK	42	/* new-req: has ack */
#define	TCPE_NQSYN	43	/* new-req: has no syn */
#define	TCPE_PASSFIN	44	/* want to send non-existing stuff */

/*
 * tcp macros
 * t_wcc (write count) is exact
 */
#define t_rhiw(tdp)	tdp->td_rdq->q_hiwat
#define t_rwin(tdp)	(tdp->td_rdq ? (t_rhiw(tdp) - tdp->td_rcount) : 0)
#define t_wcc(tdp)	tdp->td_wcount

/*
 * TCP FSM state definitions.
 * Per RFC793, September, 1981.
 */

#define	TCP_NSTATES	11

#define	TV_CLOSED		0	/* closed */
#define	TV_LISTEN		1	/* listening for connection */
#define	TV_SYN_SENT		2	/* active, have sent syn */
#define	TV_SYN_RCVD		3	/* have send and received syn */
/* states < TV_ESTABLISHED are those where connections not established */
#define	TV_ESTABLISHED		4	/* established */
#define	TV_CLOSE_WAIT		5	/* rcvd fin, waiting for close */
/* states > TV_CLOSE_WAIT are those where user has closed */
#define	TV_FIN_W1		6	/* have closed, sent fin */
#define	TV_CLOSING		7	/* closed xchd FIN; await FIN ACK */
#define	TV_LAST_ACK		8	/* had fin and close; await FIN ACK */
/* states > TV_CLOSE_WAIT && < TV_FIN_W2 await ACK of FIN */
#define	TV_FIN_W2		9	/* have closed, fin is acked */
#define	TV_TIME_WAIT		10	/* in 2*msl quiet wait after close */

#define	TV_HAVERCVDSYN(s)	((s) >= TV_SYN_RCVD)
#define	TV_HAVERCVDFIN(s)	((s) == TV_TIME_WAIT || (s) == TV_CLOSING || \
					(s) == TV_LAST_ACK|| (s) == TV_CLOSE_WAIT)
#define	TV_CANTRCVMORE(s)	((s)==TV_CLOSE_WAIT || (s)==TV_LAST_ACK || \
				 (s)==TV_CLOSING || (s)==TV_TIME_WAIT)

extern mblk_t *tv_popen(), *bp_dupall();
extern struct tcpcb *tn_connreq();

extern int tcp_run_serv, tcp_need_timeout;
#define TCP_LOCK_TIMEOUT() {int s; s = splhi(); tcp_run_serv = 1; splx(s);}
#define TCP_RUN_TIMEOUT() {int s = splhi(); \
		tcp_run_serv = 0; \
		if (tcp_need_timeout) { \
			(void) splx (s); tcptimeout(); \
		}else (void) splx (s); }
#define	WQREPLY(q, bp)	{q = RD(q); putnext(q, bp);}

#define	TCP_ADDRLEN		16 /* Long + short sizeof does not work */
#ifndef TCP_BINDADDRLEN
#define	TCP_BINDADDRLEN		TCP_ADDRLEN
#endif /* TCP_BINDADDRLEN */
#define	TCP_BINDACKLEN		sizeof(struct T_bind_ack) + TCP_ADDRLEN
#define	TCP_MAXCONIND		4	/* # of outstanding indications */

#define	TCP_CONNCONRESOFFSET	sizeof(struct T_conn_con)
#define	TCP_CONNCONRESLEN	TCP_ADDRLEN
#define	TCP_CONNCONLEN		TCP_CONNCONRESOFFSET + TCP_CONNCONRESLEN
#define	TCP_CONNINDSRCOFFSET	sizeof(struct T_conn_ind)
#define	TCP_CONNINDSRCLEN	TCP_ADDRLEN
#define	TCP_CONNINDLEN		TCP_CONNINDSRCOFFSET + TCP_CONNINDSRCLEN

#define TCPPORT_RESERVED	1024	/* Non root UID */

struct	tcpstat {
	int	tcps_badsum;
#ifdef SHORT_IDENT
	int	Tcps_badoff;
#define tcps_badoff	Tcps_badoff	/*Compiler bug */
#else
	int	tcps_badoff;
#endif /* SHORT_IDENT */
	int	tcps_hdrops;
#ifdef SHORT_IDENT
	int	TCps_badsegs;
#define tcps_badsegs	TCps_badsegs	/*Compiler */
#else
	int	tcps_badsegs;
#endif /* SHORT_IDENT */
	int	tcps_unack;
};
#ifdef	TCPSTATES
char *tcpstates[] = {
	"CLOSED",	"LISTEN",	"SYN_SENT",	"SYN_RCVD",
	"ESTABLISHED",	"CLOSE_WAIT",	"FIN_WAIT_1",	"CLOSING",
	"LAST_ACK",	"FIN_WAIT_2",	"TIME_WAIT",
};
#endif
